"""
Module providing methods for detecting adversarial samples under a common interface.
"""
from art.detection.detector import BinaryInputDetector, BinaryActivationDetector
from art.detection.subsetscanning.detector import SubsetScanningDetector
